#include <iostream>
#include <fstream> // For reading files
#include <vector>
#include <random>
using namespace std;

int rand_int(int min, int max)
{
	// Use time for setting the intial seed of the random generator
	int seed = time(NULL);
	static default_random_engine generator(seed);
	uniform_int_distribution<int> distribution(min, max - 1);
	return distribution(generator);
}

void usage(char *argv[])
{
	// Print the arguments this program expects from command line
	cout << "Usage " << argv[0] << " <F> <K> <I> <T> <R>" << endl;
	cout << "F: Path to the file" << endl;
	cout << "K: Number of clusters" << endl;
	cout << "I: Maximum number of iterations" << endl;
	cout << "T: Convergance threshold" << endl;
	cout << "R: Number of executions (repeat)" << endl;
}

void read_points(fstream &file, double **points, int N, int D)
{
	// Read points from the open file
	for (int i = 0; i < N; i++) {
		for (int j = 0; j < D; j++) {
			double tmp;
			file >> tmp;
			points[i][j] = tmp;
		}
	}
}

void print_data(double **points, int N, int D)
{
	// Print points to the terminal
	cout << "Data: " << endl;
	for (int i = 0; i < N; i++) {
		for (int j = 0; j < D; j++) {
			cout << points[i][j] << " ";
		}
		cout << endl;
	}
}

void select_centers_uniformly(double **points, double **centers, int N, int D, int K)
{
	int c = 0;
	for (int i = 0; i < K; i++) {
		int random_index = rand_int(0, N);
		// cout << "center: " << c << " random: " << random_index << endl;
		for (int j = 0; j < D; j++) {
			// select the random point as the center of cluster
			centers[c][j] = points[random_index][j];
		}
		c++;
	}
}

void print_centers(double **centers, int K, int D)
{
	for (int i = 0; i < K; i++) {
		for (int j = 0; j < D; j++) {
			cout << centers[i][j] << " ";
		}
		cout << endl;
	}
}

int main(int argc, char *argv[])
{
	const int count_arguments = 5;
	char *file_path;
	int K;
	int iters;
	double threshold;
	int R;
	fstream file;
	int count_points;
	int point_dimenssion;
	double **points;
	double **centers;

	// Check if the number of arguments is valid
	if (argc <= count_arguments) {
		cout << "Error: expecting " << count_arguments <<
			" arguments but " << argc - 1 << " have been provided" << endl;
		usage(argv);
		return 1;
	}

	// Check that arguments are valid
	file_path = argv[1];
	K = atoi(argv[2]);
	iters = atoi(argv[3]);
	threshold = atof(argv[4]);
	R = atoi(argv[5]);
	if (K == 0) {
		cout << "Unexpected value for number of clusters (K)" << endl;
		return 1;
	}
	if (iters == 0) {
		cout << "Unexpected value for number of iterations (I)" << endl;
		return 1;
	}
	if (R == 0) {
		cout << "Unexpected value for number of repeats (R)" << endl;
		return 1;
	}

	// Just for test: see the input parameters
	// cout << "Parameters:" << endl;
	// cout << "K = " << K << " T = " << threshold << " I = " << iters << " R = " << R << endl;

	// Open file for reading
	file.open(file_path, fstream::in);

	// First line has N and D
	file >> count_points;
	file >> point_dimenssion;
	// cout << "N = " << count_points << " D = " << point_dimenssion << endl;

	// Allocate memory (2d array) for points
	points = new double *[count_points];
	if (!points) {
		cout << "Failed to allocate memory for points!" << endl;
		return 1;
	}
	for (int i = 0; i < count_points; i++) {
		points[i] = new double[point_dimenssion];
		if (!points[i]) {
			cout << "Failed to allocate memory for points!" << endl;
			return 1;
		}
	}

	// Allocate memory (2d array) for center of clusters
	centers = new double *[K];
	if (!centers) {
		cout << "Failed to allocate memory!" << endl;
		return 1;
	}
	for (int i = 0; i < K; i++) {
		centers[i] = new double[point_dimenssion];
		if (!centers[i]) {
			cout << "Failed to allocate memory!" << endl;
			return 1;
		}
	}

	// Read points from the file
	// cout << "Reading file: " << file_path << endl;
	read_points(file, points, count_points, point_dimenssion);

	// Close the file
	file.close();

	// Just for testing that data was read correctly from the file
	// print_data(points, count_points, point_dimenssion);

	// Randomly and with uniform distribution selet the center of each cluster
	select_centers_uniformly(points, centers, count_points, point_dimenssion, K);

	// Print the selected centers to the terminal
	print_centers(centers, K, point_dimenssion);

	// Free the memory
	for (int i = 0; i < count_points; i++)
		delete[] points[i];
	delete[] points;
	for (int i = 0; i < K; i++)
		delete[] centers[i];
	delete[] centers;
	return 0;
}
