#include <iostream>
#include <fstream> // For reading files
#include <vector>
#include <random>
using namespace std;

int rand_int(int min, int max)
{
	// Generate a random integer in the range of [min, max)
	// Use time for setting the intial seed of the random generator
	int seed = time(NULL);
	static default_random_engine generator(seed);
	uniform_int_distribution<int> distribution(min, max - 1);
	return distribution(generator);
}

void usage(char *argv[])
{
	// Print the arguments this program expects from command line
	cout << "Usage " << argv[0] << " <F> <K> <I> <T> <R>" << endl;
	cout << "F: Path to the file" << endl;
	cout << "K: Number of clusters" << endl;
	cout << "I: Maximum number of iterations" << endl;
	cout << "T: Convergance threshold" << endl;
	cout << "R: Number of executions (repeat)" << endl;
}

void read_points(fstream &file, double **points, int N, int D)
{
	// Read points from the open file
	for (int i = 0; i < N; i++) {
		for (int j = 0; j < D; j++) {
			double tmp;
			file >> tmp;
			points[i][j] = tmp;
		}
	}
}

void print_data(double **points, int N, int D)
{
	// Print points to the terminal
	cout << "Data: " << endl;
	for (int i = 0; i < N; i++) {
		for (int j = 0; j < D; j++) {
			cout << points[i][j] << " ";
		}
		cout << endl;
	}
}

void select_centers_uniformly(double **points, double **centers, int N, int D, int K)
{
	int c = 0;
	for (int i = 0; i < K; i++) {
		int random_index = rand_int(0, N);
		// cout << "center: " << c << " random: " << random_index << endl;
		for (int j = 0; j < D; j++) {
			// select the random point as the center of cluster
			centers[c][j] = points[random_index][j];
		}
		c++;
	}
}

void print_centers(double **centers, int K, int D)
{
	for (int i = 0; i < K; i++) {
		for (int j = 0; j < D; j++) {
			cout << centers[i][j] << " ";
		}
		cout << endl;
	}
}

double euclidean_distance(double *p1, double *p2, int D)
{
	// Calculate distance between P1 and P2
	double dist = 0;
	double tmp;
	for (int i = 0; i < D; i++) {
		tmp = p1[i] - p2[i];
		tmp = tmp * tmp;
		dist += tmp;
	}
	return dist;
}

/*
 * @params p1: data point
 * @params centers: list of center point for each cluster
 * @params K: number of clusters
 * @params D: dimensions of each point
 * @params _min_dist: write the distance to the nearest center value in this variable
 * @returns the index of nearest cluster
 * */
int assign_cluster(double *p1, double **centers, int K , int D, double *_min_dist)
{
	// Find nearest center point to p1 and return its index

	int min_index = -1;
	double min_dist = 1 << 30; // Select a very large number
	double dist;
	for (int i = 0; i < K; i++) {
		dist = euclidean_distance(p1, centers[i], D);
		if (dist < min_dist) {
			// This cluster is the closest we have observed until now
			min_dist = dist;
			min_index = i;
		}
	}
	*_min_dist = min_dist;
	return min_index;
}

/*
 * @params points: all the points (2d array with dimensions N x D)
 * @params cluster: a 1d array indicating the cluster of each point
 *                  points[i] belongs to the cluster[i].
 * @params centers: cluster center values (2d array with dimensions K x D)
 *                  This array will be updated.
 * @param N: number of points
 * @param K: number of clusters
 * @param D: dimension of a point
 * */
void update_center_value(double **points, int *cluster, double **centers, int N, int K, int D)
{
	// How points are in each cluster
	int *size_cluster = new int[K];
	int index;

	// Initialize variables
	for (int i = 0; i < K; i++) {
		// initialize size of the i-th cluster to zero
		size_cluster[i] = 0;
		for (int j = 0; j < D; j++) {
			// clear the center value to calculate the new summation
			centers[i][j] = 0;
		}
	}

	for (int p = 0; p < N; p++) {
		index = cluster[p];
		// Sum all the points in the cluster with number = index;
		for (int j = 0; j < D; j++) {
			centers[index][j] += points[p][j];
		}
		size_cluster[index] += 1;
	}

	// Divide by count to acquire the average
	for (int i = 0; i < K; i++) {
		// Avoid division by zero
		if (size_cluster[i] == 0)
			continue;

		for (int j = 0; j < D; j++) {
			centers[i][j] /= size_cluster[i];
		}
	}

	delete[] size_cluster;
}

int run_kmeans(double **points, int *cluster, double **centers, int N, int K, int D, int I, double T)
{
	double sse = -1;
	double prev_sse = -1;
	double sse_delta;
	int cluster_index;
	double min_dist;

	// Iterate until convergance or reaching the limit
	for (int i = 0; i < I; i++) {
		prev_sse = sse;
		sse = 0;
		// Assign each point to the nearest cluster
		for (int p = 0; p < N; p++) {
			min_dist = 0;
			cluster_index = assign_cluster(points[p], centers, K, D, &min_dist);
			if (cluster_index == -1) {
				cout << "Error: could not find a suitable cluster (BUG)" << endl;
				return 1;
			}
			cluster[p] = cluster_index;
			// cout << "min dist: " << min_dist << endl;
			sse += min_dist;
		}

		if (i > 0) {
			cout << "Iteration #" << i << ": SSE = " << sse << endl;
			sse_delta = prev_sse - sse;
			if (sse_delta < 0) {
				cout << "Warning: value of sse has increased!" << endl;
			}
			if ((sse_delta / prev_sse) <= T) {
				// Algorithm has converged
				break;
			}
		}

		// Update center value of each cluster
		update_center_value(points, cluster, centers, N, K, D);
	}
	return 0;
}

int main(int argc, char *argv[])
{
	const int count_arguments = 5;
	char *file_path;
	int K;
	int iters;
	double threshold;
	int R;
	fstream file;
	int count_points;
	int point_dimenssion;
	double **points;
	double **centers;
	int *cluster;
	int ret;

	// Check if the number of arguments is valid
	if (argc <= count_arguments) {
		cout << "Error: expecting " << count_arguments <<
			" arguments but " << argc - 1 << " have been provided" << endl;
		usage(argv);
		return 1;
	}

	// Check that arguments are valid
	file_path = argv[1];
	K = atoi(argv[2]);
	iters = atoi(argv[3]);
	threshold = atof(argv[4]);
	R = atoi(argv[5]);
	if (K == 0) {
		cout << "Unexpected value for number of clusters (K)" << endl;
		return 1;
	}
	if (iters == 0) {
		cout << "Unexpected value for number of iterations (I)" << endl;
		return 1;
	}
	if (R == 0) {
		cout << "Unexpected value for number of repeats (R)" << endl;
		return 1;
	}

	// Just for test: see the input parameters
	// cout << "Parameters:" << endl;
	// cout << "K = " << K << " T = " << threshold << " I = " << iters << " R = " << R << endl;

	// Open file for reading
	file.open(file_path, fstream::in);

	// First line has N and D
	file >> count_points;
	file >> point_dimenssion;
	// cout << "N = " << count_points << " D = " << point_dimenssion << endl;

	// Allocate memory (2d array) for points
	points = new double *[count_points];
	if (!points) {
		cout << "Failed to allocate memory for points!" << endl;
		return 1;
	}
	for (int i = 0; i < count_points; i++) {
		points[i] = new double[point_dimenssion];
		if (!points[i]) {
			cout << "Failed to allocate memory for points!" << endl;
			return 1;
		}
	}

	// Allocate memory (2d array) for center of clusters
	centers = new double *[K];
	if (!centers) {
		cout << "Failed to allocate memory!" << endl;
		return 1;
	}
	for (int i = 0; i < K; i++) {
		centers[i] = new double[point_dimenssion];
		if (!centers[i]) {
			cout << "Failed to allocate memory!" << endl;
			return 1;
		}
	}

	// Allocate memory for the flags indicating each point belongs to which cluster
	cluster = new int[count_points];
	if (!cluster) {
		cout << "Failed to allocate memory!" << endl;
		return 1;
	}

	// Read points from the file
	// cout << "Reading file: " << file_path << endl;
	read_points(file, points, count_points, point_dimenssion);

	// Close the file
	file.close();

	// Just for testing that data was read correctly from the file
	// print_data(points, count_points, point_dimenssion);

	// Run the algorithm multiple times
	for (int r = 0; r < R; r++) {
		cout << "Run #" << r << endl;
		cout << "----------------------------------------------------" << endl;

		// Randomly and with uniform distribution selet the center of each cluster
		select_centers_uniformly(points, centers, count_points, point_dimenssion, K);

		// Print the selected centers to the terminal
		// print_centers(centers, K, point_dimenssion);

		ret = run_kmeans(points, cluster, centers, count_points, K, point_dimenssion, iters, threshold);
		if (ret != 0) {
			// An error has happend
			return ret;
		}
		cout << endl;
	}

	// Free the memory
	for (int i = 0; i < count_points; i++)
		delete[] points[i];
	delete[] points;
	for (int i = 0; i < K; i++)
		delete[] centers[i];
	delete[] centers;
	delete[] cluster;
	return 0;
}
